.equ IP_VER         = 0x00
.equ IP_TOS         = 0x01
.equ IP_LEN         = 0x02
.equ IP_ID          = 0x04
.equ IP_FRAG        = 0x06
.equ IP_TTL         = 0x08
.equ IP_PROTO       = 0x09
.equ IP_CKSUM       = 0x0a
.equ IP_SRC         = 0x0c
.equ IP_DST         = 0x10
.equ IP_DATA        = 0x14

.equ PROTO_ICMP     = 1
.equ PROTO_TCP      = 6
.equ PROTO_UDP      = 17

.equ ICMP_TYPE      = 0x00
.equ ICMP_CODE      = 0x01
.equ ICMP_CKSUM     = 0x02

.equ UDP_SRCPORT    = 0x00
.equ UDP_DSTPORT    = 0x02
.equ UDP_LEN        = 0x04
.equ UDP_CKSUM      = 0x06
.equ UDP_DATA       = 0x08

; Transmit an IP packet via SLIP.
; IN    Z = mem_alloc:ed buffer with packet header + data
; OUT   r1:r0 = buffer (same as Z on entry)
; r16, r18, r19 destroyed
ip_tx:
    MUTEX_WAIT r16,MUTEX_IPTX
    movw r0,zl                  ; r1:r0 = start of buffer
    ld r19,-z
    andi r19,0x7f
    ld r18,-z
    subi r18,4
    sbci r19,0                  ; r19:r18 = size of buffer
    adiw zl,2
ip_tx_byte:
    subi r18,1
    sbci r19,0
    brcs ip_tx_done
    ld r16,z+
    call slip_tx
    rjmp ip_tx_byte
ip_tx_done:
    call slip_end_packet
    MUTEX_LEAVE r16,MUTEX_IPTX
    ret

; Receive an IP packet via SLIP.
; OUT   Z = mem_alloc:ed buffer with packet header + data
;       ZF = 0 on error, 1 if successful
; r0-r4, r12-r21 destroyed
ip_rx:
    call slip_rx
    brne ip_rx_abort
    brcs ip_rx_end
    mov r12,r16                 ; r12 = version | header length
    cpi r16,0x45
    brne ip_rx_abort            ; abort unless IPv4 with 20-byte header
    call slip_rx
    brne ip_rx_abort
    brcs ip_rx_end
    mov r13,r16                 ; r13 = TOS
    call slip_rx
    brne ip_rx_abort
    brcs ip_rx_end
    mov r14,r16                 ; r14 = total length (MSB)
    call slip_rx
    brne ip_rx_abort
    brcs ip_rx_end
    mov r15,r16                 ; r15 = total length (LSB)
    mov r17,r14
    movw r18,r16                ; r17:r16 = r19:r18 = total length
    subi r18,0x41
    sbci r19,0x02
    brcc ip_rx_abort            ; abort if size > 576 bytes
    call mem_alloc
    mov r0,zl
    or r0,zh
    breq ip_rx_abort            ; abort if mem_alloc failed
    movw r0,zl                  ; r1:r0 = pointer to buffer
    st z+,r12
    st z+,r13
    st z+,r14
    st z+,r15                   ; store first four bytes
    mov r18,r15
    mov r19,r14                 ; r19:r18 = number of bytes to receive
    subi r18,5                  ; subtract 5, 4 for the ones received and
    sbci r19,0                  ; 1 since we copy one more before looping
ip_rx_byte:
    call slip_rx
    brne ip_rx_free_abort       ; free buffer and abort on error
    brcs ip_rx_free             ; or if we receive an END byte
    st z+,r16                   ; if everything is OK, store the byte
    subi r18,1
    sbci r19,0
    brcc ip_rx_byte
    movw zl,r0                  ; Z = buffer
    call slip_rx                ; this should be an END byte
    brne ip_rx_free_abort
    brcc ip_rx_free_abort       ; if it's not, abort
    ret                         ; if it is, return with ZF = 1 (success)
ip_rx_free:
    call mem_free
    rjmp ip_rx_end
ip_rx_free_abort:
    call mem_free
ip_rx_abort:
    call slip_rx                ; read until end of packet
    brne ip_rx_abort
    brcc ip_rx_abort
ip_rx_end:
    clz                         ; ZF = 0 (error)
    ret

; Compute the one's complement sum of 16-bit MSB words in the data block
; Note that IP checksums use the one's complement of this sum.
; IN    Z = data block
;       r17:r16 = length of block in bytes
;       r1:r0 = initial sum (little-endian, typically 0)
; OUT   r1:r0 = checksum (little-endian)
;       Z = old Z + old r17:r16
; r2, r3, r16, r17, destroyed
ip_ocsum:
    subi r16,1
    sbci r17,0                  ; r17:r16--
    sbrc r17,7                  ; if result is negative
    ret                         ; return with sum in r1:r0
    ld r3,z+                    ; otherwise, load MSB
    subi r16,1
    sbci r17,0                  ; r17:r16--
    sbrc r17,7                  ; if result is negative
    clr r2                      ; LSB is padded with a 0 byte
    sbrs r17,7                  ; otherwise
    ld r2,z+                    ; load LSB
    add r0,r2
    adc r1,r3                   ; r1:r0 += r3:r2
    brcc ip_ocsum               ; if no carry, continue
    inc r0
    brne ip_ocsum
    inc r1
    rjmp ip_ocsum


; Create an IP header in an already allocated buffer.
; Before calling this, the protocol byte as well as the source and
; destination addresses should be set.
; IN    Z = packet buffer
; r0, r1, r2, r3, r16, r17 destroyed
ip_create_header:
    clr r17
    ldi r16,0x45
    st z,r16                    ; IPv4, 20-byte header
    std z+IP_TOS,r17            ; no TOS options
    std z+IP_ID+0,r17
    std z+IP_ID+1,r17           ; ID field = 0
    ldi r16,0x40
    std z+IP_FRAG+0,r16         ; don't fragment
    std z+IP_FRAG+1,r17         ; fragment offset = 0
    std z+IP_TTL,r16            ; TTL = 64
    std z+IP_CKSUM+0,r17
    std z+IP_CKSUM+1,r17        ; checksum field = 0 for now

    clr r0
    clr r1                      ; initial sum = 0
    ldi r16,20                  ; r17:r16 = 20 (length of header)
    rcall ip_ocsum
    sbiw zl,20
    com r0
    com r1                      ; r1:r0 = IP header checksum
    std z+IP_CKSUM+0,r1
    std z+IP_CKSUM+1,r0         ; store into buffer
    ret

; Swap the source and destination addresses of an IP packet
; IN    Z = packet buffer
; r0, r1, r16 destroyed
ip_swap_adr:
    ldi r16,4
    adiw zl,IP_SRC
ip_swap_adr_loop:
    ldd r1,z+4
    ld r0,z
    std z+4,r0
    st z+,r1
    dec r16
    brne ip_swap_adr_loop
    sbiw zl,IP_SRC+4
    ret

