; Memory block structure
;
; 00    word    n, size of block  in bytes (bit 15 = alloced flag)
; 02    word    previous block in free list
; 04    word    next block in free list
; ...   ...     ...
; n-2   word    n, size of block in bytes (bit 15 = alloced flag)

; Initialize memory allocator.
; r16, r17 destroyed
mem_init:
    ldi r16,low(FREE_RAM_END-FREE_RAM_START)
    sts FREE_RAM_START,r16
    ldi r17,high(FREE_RAM_END-FREE_RAM_START)
    sts FREE_RAM_START+1,r17
    sts FREE_RAM_END-2,r16
    sts FREE_RAM_END-1,r17
    clr r16
    sts FREE_RAM_START+2,r16
    sts FREE_RAM_START+3,r16
    sts FREE_RAM_START+4,r16
    sts FREE_RAM_START+5,r16
    ldi r16,low(FREE_RAM_START)
    ldi r17,high(FREE_RAM_START)
    sts MEM_FREELISTL,r16
    sts MEM_FREELISTH,r17
    ret

; IN    r17:16 = requested size of block, in bytes
; OUT   Z = block data area (may be up to 4 bytes larger than requested)
;       Z = 0 on error
; r0, r1, r2, r3, r4, r16, r17, r18, r19 destroyed
mem_alloc:
    rcall mem_wait
    subi r16,-4
    sbci r17,0xff               ; r17:16 += 4 (need space for metadata)
    lds zl,MEM_FREELISTL
    lds zh,MEM_FREELISTH        ; Z points to element of freelist
mem_alloc_check:
    mov r0,zl
    or r0,zh
    breq mem_alloc_ret          ; abort if Z = 0
    ld r18,z
    ldd r19,z+1                 ; r19:r18 = size of block (alloc flag = 0)
    sub r18,r16
    sbc r19,r17                 ; block size -= requested size
    brcs mem_alloc_next         ; If the block size is smaller, try next
    brne mem_alloc_split        ; >=256b, block is large enough to split
    cpi r18,4
    brcc mem_alloc_split        ; >=4b, block is large enough to split
mem_alloc_whole:                ; block too small, allocate all of it
    ld r16,z
    ldd r17,z+1                 ; r17:r16 = size of block (alloc flag = 0)
    add zl,r16
    adc zh,r17
    ld r18,-z
    ori r18,0x80                ; r18 = MSB of tail size field | 0x80
    st z,r18                    ; set alloc flag (tail)
    sub zl,r16
    sbc zh,r17
    std z+1,r18                 ; set alloc flag (head)
    movw r18,zl
    rcall mem_delete            ; delete it from free list
    movw zl,r18
    adiw zl,2                   ; Z = start of data area
mem_alloc_ret:
    rcall mem_leave
    ret
mem_alloc_next:
    ldd r0,z+4
    ldd zh,z+5
    mov zl,r0                   ; Z = address of next block
    rjmp mem_alloc_check
mem_alloc_split:                ; create new block of size r19:r18
    push zl
    push zh
    rcall mem_delete            ; first, delete the large block
    pop zh
    pop zl
    movw r0,zl
    st z,r16
    ori r17,0x80
    std z+1,r17                 ; store size of alloced block (head)
    andi r17,0x7f
    add zl,r16
    adc zh,r17
    ori r17,0x80
    st -z,r17
    st -z,r16                   ; store size of alloced block (tail)
    adiw zl,2
    st z,r18
    std z+1,r19                 ; store size of free block (head)
    rcall mem_insert
    movw zl,r0
    adiw zl,2
    rjmp mem_leave

; Delete a block from the free list.
; IN    Z = block
; r0, r1, r2, r3, r4, zl, zh destroyed.
mem_delete:
    ldd r0,z+2
    ldd r1,z+3                  ; r1:r0 = block->prev
    ldd r2,z+4
    ldd r3,z+5                  ; r3:r2 = block->next
    mov r4,r0
    or r4,r1
    breq mem_delete_head
    movw zl,r0                  ; block->prev != null
    std z+4,r2
    std z+5,r3                  ; block->prev->next = block->next
    rjmp mem_delete_next
mem_delete_head:                ; block->prev == null, delete head of list
    sts MEM_FREELISTL,r2
    sts MEM_FREELISTH,r3        ; freelist = block->next
mem_delete_next:
    mov r4,r2
    or r4,r3
    breq mem_delete_tail
    movw zl,r2                  ; block->next != null
    std z+2,r0
    std z+3,r1                  ; block->next->prev = block->prev
mem_delete_tail:
    ret

mem_wait:
    MUTEX_WAIT r18,MUTEX_MEM
    ret

mem_leave:
    MUTEX_LEAVE r16,MUTEX_MEM
    ret

; Allocate a memory block.
; Insert a block into the free list.
; Also write the size word at the end of the block.
; IN    Z = block to insert
;       r19:r18 = size of block
; r2, r3, r4, r16, r17, zl, zh destroyed
mem_insert:
    lds r2,MEM_FREELISTL
    lds r3,MEM_FREELISTH        ; r3:r2 = head of free list
    clr r4
    std z+2,r4
    std z+3,r4                  ; new->prev = null (first in free list)
    std z+4,r2
    std z+5,r3                  ; new->next = old head of free list
    sts MEM_FREELISTL,zl
    sts MEM_FREELISTH,zh
    movw r16,zl
    add zl,r18
    adc zh,r19
    st -z,r19
    st -z,r18                   ; store size of free block (tail)
    movw zl,r2
    or r2,r3
    breq mem_insert_nonext      ; return if new->next == null
    std z+2,r16
    std z+3,r17                 ; new->next->prev = new
mem_insert_nonext:
    ret

; Free a previously allocated (by mem_alloc) memory block.
; IN    Z = block data area
; r0, r1, r2, r3, r4, r16, r17, r18, r19, r20, r21, zl, zh destroyed
mem_free:
    rcall mem_wait
    sbiw zl,2                   ; Z = start of block
    movw r18,zl                 ; r19:r18 = start of merged block
    cpi zl,low(FREE_RAM_START)
    brne mem_free_notfirst
    cpi zh,high(FREE_RAM_START)
    breq mem_free_noprev        ; is this the first block in RAM?
mem_free_notfirst:              ; no, check if the preceding block is free
    ld r17,-z
    ld r16,-z                   ; r17:r16 = size of preceding block
    adiw zl,2
    cpi r17,0x80                ; is the preceding block free?
    brcc mem_free_noprev        ; no, so don't merge with it
    sub zl,r16
    sbc zh,r17                  ; Z = preceding block
    movw r18,zl                 ; r19:18 = start of merged block
    rcall mem_delete            ; delete it from the free list
    movw zl,r18
    add zl,r16
    adc zh,r17                  ; Z = original block
mem_free_noprev:
    ld r16,z
    ldd r17,z+1                 ; r17:16 = size of block | 0x8000
    andi r17,0x7f               ; clear alloc flag
    std z+1,r17
    add zl,r16
    adc zh,r17                  ; Z = following block (or end of RAM)
    movw r16,zl                 ; r17:r16 = end of merged block
    cpi zl,low(FREE_RAM_END)
    brne mem_free_notlast
    cpi zh,high(FREE_RAM_END)
    breq mem_free_nonext        ; is this the last block in RAM?
mem_free_notlast:               ; no, check if the following block is free
    ld r20,z
    ldd r21,z+1                 ; r21:r20 = size of following block
    cpi r21,0x80                ; is the following block free?
    brcc mem_free_nonext        ; no, so don't merge with it
    add r16,r20
    adc r17,r21                 ; add this block's size to merged block's
    rcall mem_delete
mem_free_nonext:
    movw zl,r18                 ; Z = start of merged block
    sub r16,zl
    sbc r17,zh                  ; r17:r16 = size of merged block
    st z,r16
    std z+1,r17                 ; store size of merged block (head)
    movw r18,r16                ; r19:r18 = size of block
    rcall mem_insert            ; insert block into free list & store tail
    rjmp mem_leave

