; Notes on timing.
;
; Maximum UART speed is 115200 baud, or 11520 bytes received per second.
; At 7.3728 MHz, this means 640 cycles between UART RXC interrupts.
;
; Timer 0 is set to CLK/64, which means one overflow interrupt is
; generated every 16384 cycles, at a frequency of 450 Hz.
;
; A typical task switch requires some 170 cycles, so the overhead is 1%
; Every 2^16 ticks there's a 20 cycle penalty for updating the top bytes of
; UPTIME.
; For every WAITing process in the queue there's 13 more cycles.

; Proc structure (pointed to by Y)
;
; 00    word    previous proc structure
; 02    word    next proc structure
; 04    byte    flags (bit 7 = WAIT?)
; 05    byte    SREG
; 06    word    SP
; 08    30b     r0 -- r31, except Y (r28, r29)
; 26

.equ PROC_PREV      = 0x00
.equ PROC_NEXT      = 0x02
.equ PROC_FLAGS     = 0x04
.equ PROC_SREG      = 0x05
.equ PROC_SP        = 0x06
.equ PROC_REGS      = 0x08
.equ PROC_DATA      = 0x26

; Initialize the scheduler. No process is created yet, the user should
; call sch_create and then sch_yield to launch the first process.
sch_init:
    clr yl
    sts UPTIME+0,yl
    sts UPTIME+1,yl
    sts UPTIME+2,yl
    sts UPTIME+3,yl
    sts UPTIME+4,yl
    sts MUTEX,yl                ; clear mutex bits
    sts SCH_PROCLISTL,yl
    sts SCH_PROCLISTH,yl        ; create empty list of processes
    sts SCH_NPROC,yl            ; 0 active processes
    ser yh                      ; set MSB of Y, since no process exists
    out TCNT0,yl                ; clear timer 0 counter
    ldi yl,0x03                 ; CK/64 prescaler (16384 cycle interrupt)
    out TCCR0,yl                ; start timer 0
    in yl,TIMSK
    ori yl,1<<TOIE0
    out TIMSK,yl                ; enable timer 0 overflow interrupt
    ret

; This is called internally if there is no active process.
sch_idle:
    sbrc yh,7                   ; if the MSB of Y is set, we have an
    rjmp sch_idle_nosave        ; invalid address, so skip saving
    rcall sch_saveregs
    in r0,SPL
    std y+PROC_SP+0,r0
    in r0,SPH
    std y+PROC_SP+1,r0
sch_idle_nosave:
    ser yh                      ; show that this is not a process
    sei                         ; wait for some interrupt
    rjmp PC                     ; TODO: consider some sleep mode here

sch_reti:
    reti

; sch_yield: Make the current process yield its remaining time.
; sch_wait: Like yield, but the process is also put in a WAIT state.
; r16 is destroyed.
sch_wait:
    cli
    push r16
    ldd r16,y+PROC_FLAGS
    ori r16,0x80
    std y+PROC_FLAGS,r16        ; set WAIT bit
    lds r16,SCH_NPROC
    dec r16
    sts SCH_NPROC,r16           ; one less active process
    pop r16
    breq sch_idle
sch_yield:
    cli
    sbrc yh,7                   ; if the MSB of Y is set, we have an
    rjmp sch_switch_first       ; invalid address, so skip saving
    rcall sch_saveregs
    rjmp sch_switch

int_tim0_ovf:
    sbrs yh,7                   ; if we're not in a process
    rcall sch_saveregs          ; then skip saving

    wdr                         ; watchdog reset

    lds r16,UPTIME+0
    subi r16,0xff
    sts UPTIME+0,r16
    brcs sch_switch
    lds r16,UPTIME+1
    sbci r16,0xff
    sts UPTIME+1,r16
    brcs sch_switch
    lds r16,UPTIME+2
    sbci r16,0xff
    sts UPTIME+2,r16
    lds r16,UPTIME+3
    sbci r16,0xff
    sts UPTIME+3,r16
    lds r16,UPTIME+4
    sbci r16,0xff
    sts UPTIME+4,r16            ; UPTIME++

sch_switch:
    sbrc yh,7                   ; if the MSB of Y is set, we have an
    rjmp sch_switch_first       ; invalid address
sch_switch_loop:                ; about 13 cycles per loop
    ldd r0,y+PROC_NEXT+0
    ldd yh,y+PROC_NEXT+1
    or r0,yh                    ; is this the last proc struct?
    brne sch_switch_notlast
sch_switch_first:
    lds r1,SCH_NPROC
    tst r1
    breq sch_reti
    lds r0,SCH_PROCLISTL
    lds yh,SCH_PROCLISTH        ; yes, jump back to first proc struct
sch_switch_notlast:
    mov yl,r0
    ldd r0,y+PROC_FLAGS
    tst r0                      ; is this proc in WAIT state?
    brmi sch_switch_loop        ; yes, find another one

sch_load:                       ; 73 cycles (including return)
    ldd r0,y+PROC_SP+0
    out SPL,r0
    ldd r0,y+PROC_SP+1
    out SPH,r0
    ldd r0,y+PROC_SREG
    out SREG,r0
    ldd r0,y+0x08
    ldd r1,y+0x09
    ldd r2,y+0x0a
    ldd r3,y+0x0b
    ldd r4,y+0x0c
    ldd r5,y+0x0d
    ldd r6,y+0x0e
    ldd r7,y+0x0f
    ldd r8,y+0x10
    ldd r9,y+0x11
    ldd r10,y+0x12
    ldd r11,y+0x13
    ldd r12,y+0x14
    ldd r13,y+0x15
    ldd r14,y+0x16
    ldd r15,y+0x17
    ldd r16,y+0x18
    ldd r17,y+0x19
    ldd r18,y+0x1a
    ldd r19,y+0x1b
    ldd r20,y+0x1c
    ldd r21,y+0x1d
    ldd r22,y+0x1e
    ldd r23,y+0x1f
    ldd r24,y+0x20
    ldd r25,y+0x21
    ldd r26,y+0x22
    ldd r27,y+0x23
    ldd r30,y+0x24
    ldd r31,y+0x25
    reti

; zl, zh destroyed
sch_saveregs:                   ; 75 cycles (including ret, not call)
    std y+0x08,r0
    std y+0x09,r1
    std y+0x0a,r2
    std y+0x0b,r3
    std y+0x0c,r4
    std y+0x0d,r5
    std y+0x0e,r6
    std y+0x0f,r7
    std y+0x10,r8
    std y+0x11,r9
    std y+0x12,r10
    std y+0x13,r11
    std y+0x14,r12
    std y+0x15,r13
    std y+0x16,r14
    std y+0x17,r15
    std y+0x18,r16
    std y+0x19,r17
    std y+0x1a,r18
    std y+0x1b,r19
    std y+0x1c,r20
    std y+0x1d,r21
    std y+0x1e,r22
    std y+0x1f,r23
    std y+0x20,r24
    std y+0x21,r25
    std y+0x22,r26
    std y+0x23,r27
    std y+0x24,r30
    std y+0x25,r31
    in zl,SREG
    std y+PROC_SREG,zl
    in zl,SPL
    in zh,SPH
    adiw zl,2
    std y+PROC_SP+0,zl
    std y+PROC_SP+1,zh
    ret

; Create a new process.
; IN    r17:r16 = size of heap + stack
;       r19:r18 = initial execution address
; OUT   Z = proc struct, 0 on error
; r0, r1, r2, r3, r4, r16, r17, r18, r19 destroyed
; The contents of the other registers are inherited by the new process
sch_create:
    in r0,SREG
    push r0
    cli
    subi r16,-0x26
    sbci r17,0xff               ; we also need room for the proc struct
    push r18
    push r19
    call mem_alloc              ; allocate proc struct + heap + stack area
    pop r19
    pop r18
    mov r0,zl
    or r0,zh
    breq sch_create_ret         ; abort if allocation failed
    movw r0,yl
    movw yl,zl
    rcall sch_saveregs          ; write register values to new process
    movw zl,yl
    ld r17,-y
    andi r17,0x7f
    ld r16,-y                   ; r17:r16 = size of mem_alloc block
    add yl,r16
    adc yh,r17
    sbiw yl,5                   ; Y = top of allocated area - 3
    std y+2,r18                 ; this should be stored big-endian!
    std y+1,r19                 ; store return address at top of stack
    std z+0x06,yl
    std z+0x07,yh               ; store SP of new process
    movw yl,r0
    clr r0
    std z+PROC_FLAGS,r0         ; WAIT = 0
    lds r2,SCH_PROCLISTL
    lds r3,SCH_PROCLISTH        ; r3:r2 = head of proc list
    st z,r0
    std z+1,r0                  ; proc->prev = null (first in list)
    std z+2,r2
    std z+3,r3                  ; proc->next = old head of list
    sts SCH_PROCLISTL,zl
    sts SCH_PROCLISTH,zh
    movw r0,zl
    movw zl,r2
    or r2,r3
    breq sch_create_nonext      ; skip if proc->next == null
    st z,r0
    std z+1,r1                  ; proc->next->prev = proc
sch_create_nonext:
    lds r16,SCH_NPROC
    inc r16
    sts SCH_NPROC,r16           ; one more active process
sch_create_ret:
    pop r0
    out SREG,r0
    ret

; Remove a process from the list and deallocate its memory.
; IN    Z = process
; r0, r1, r2, r3, r4, r16, r17, r18, r19, r20, r21, zl, zh destroyed
sch_kill:
    in r0,SREG
    push r0
    cli
    ld r0,z
    ldd r1,z+1                  ; r1:r0 = proc->prev
    ldd r2,z+2
    ldd r3,z+3                  ; r3:r2 = proc->next
    movw r16,zl                 ; r17:r16 = proc
    mov r4,r0
    or r4,r1
    breq sch_kill_head
    movw zl,r0                  ; proc->prev != null
    std z+2,r2
    std z+3,r3                  ; proc->prev->next = proc->next
    rjmp sch_kill_next
sch_kill_head:
    sts SCH_PROCLISTL,r2
    sts SCH_PROCLISTH,r3        ; proclist = proc->next
sch_kill_next:
    mov r4,r2
    or r4,r3
    breq sch_kill_tail
    movw zl,r2                  ; proc->next != null
    st z,r0
    std z+1,r1                  ; proc->next->prev = proc->prev
sch_kill_tail:
    pop r0
    out SREG,r0
    movw zl,r16
    ldd r16,z+PROC_FLAGS
    push r16
    rcall mem_free
    pop r16
    sbrc r16,7                  ; if this process was sleeping
    ret                         ; return without changing process count
    lds r16,SCH_NPROC
    dec r16
    sts SCH_NPROC,r16           ; one process less
    brne sch_kill_ret           ; did we run out of active processes?
    jmp sch_idle                ; yes, idle
sch_kill_ret:
    ret                         ; no, just return

; Kill the current process. Does not return.
sch_die:
    cli
    movw zl,yl
    ser yh
    rcall sch_kill
    rjmp sch_yield


; Delay for (at least) r4:r3:r2:r1:r0 ticks.
; r0-r5 destroyed
sch_delay:
    lds r5,UPTIME+0
    add r0,r5
    lds r5,UPTIME+1
    adc r1,r5
    lds r5,UPTIME+2
    adc r2,r5
    lds r5,UPTIME+2
    adc r3,r5
    lds r5,UPTIME+2
    adc r4,r5
sch_delay_wait:
    call sch_yield
    lds r5,UPTIME+0
    cp r0,r5
    lds r5,UPTIME+1
    cpc r1,r5
    lds r5,UPTIME+2
    cpc r2,r5
    lds r5,UPTIME+3
    cpc r3,r5
    lds r5,UPTIME+4
    cpc r4,r5
    brcc sch_delay_wait
    ret

