; TODO: more input buffering

.equ SLIP_END       = 0xc0
.equ SLIP_ESC       = 0xdb
.equ SLIP_ESC_END   = 0xdc
.equ SLIP_ESC_ESC   = 0xdd

; Transmit one byte using the SLIP encoding.
; IN    r16 = data byte
; r16 is destroyed
slip_tx:
    cpi r16,SLIP_END
    breq slip_tx_end
    cpi r16,SLIP_ESC
    breq slip_tx_esc
    rjmp uart_tx
slip_tx_end:
    ldi r16,SLIP_ESC
    rcall uart_tx
    ldi r16,SLIP_ESC_END
    rjmp uart_tx
slip_tx_esc:
;   ldi r16,SLIP_ESC
    rcall uart_tx
    ldi r16,SLIP_ESC_ESC
    rjmp uart_tx

; Transmit SLIP END, to signal end of packet
; r16 is destroyed
slip_end_packet:
    ldi r16,SLIP_END
    rjmp uart_tx

; Receive one byte via SLIP.
; OUT   r16 = data byte (if CF = 0 and ZF = 1)
;       ZF = 0 on error
;       CF = 1 and ZF = 1 if we're at the end of a packet (r16 = SLIP_END)
slip_rx:
    rcall uart_rx
    brne slip_rx_err            ; on reception error, abort packet
    cpi r16,SLIP_ESC
    breq slip_rx_esc
    cpi r16,SLIP_END
    breq slip_rx_end
    clc
    sez                         ; ZF = 1, CF = 0 (normal byte)
    ret
slip_rx_end:
    sec                         ; ZF = 1, CF = 1 (end of packet)
    ret
slip_rx_esc:
    rcall uart_rx
    brne slip_rx_err
    cpi r16,SLIP_ESC_END
    breq slip_rx_esc_end
    cpi r16,SLIP_ESC_ESC
    breq slip_rx_esc_esc
    ; unknown escape sequence counts as error, so fall through here
slip_rx_err:
    clz                         ; ZF = 0 (error)
    ret
slip_rx_esc_end:
    ldi r16,SLIP_END
    ret                         ; ZF = 1, CF = 0 (normal byte)
slip_rx_esc_esc:
    ldi r16,SLIP_ESC
    ret                         ; ZF = 1, CF = 0 (normal byte)

; Initialize the UART. Enable the transmitter.
; r16 is destroyed
uart_init:
    clr r16
    out UBRRHI,r16
    EEPROM_READ_IMM EE_BAUDRATE,r16
    out UBRR,r16                ; set baud rate
    ldi r16,(1<<TXEN)|(1<<RXEN)
    out UCSRB,r16               ; enable TX, RX, no interrupts yet
    ret

; uart_tx: Transmit one byte, wait for buffer to clear if necessary.
; It is important that only one process at a time uses this.
; IN    r16 = data byte
uart_tx_wait:
    sts UART_TX_PROCL,yl        ; register our process as the one using
    sts UART_TX_PROCH,yh        ; the UART transmitter
    cli
    sbi UCSRB,UDRIE             ; enable Data Register Empty interrupt
    call sch_wait               ; halt process until interrupt
uart_tx:
    cbi LED_PORT,1
    sbis UCSRA,UDRE             ; wait until the UART is ready to transmit
    rjmp uart_tx_wait
    out UDR,r16                 ; write data byte to output buffer
    sbi LED_PORT,1
    ret

; Data Register Empty interrupt handler.
; Switch to the process whole proc structure is pointed to by UART_TX_PROC
int_uart_dre:
    sbrc yh,7                   ; did we interrupt a process?
    rjmp int_uart_dre_nosave    ; if yes, save its state

    call sch_saveregs
    in r0,SPL
    std y+0x06,r0
    in r0,SPH
    std y+0x07,r0

int_uart_dre_nosave:
    cbi UCSRB,UDRIE             ; disable Data Register Empty interrupt

    lds yl,UART_TX_PROCL
    lds yh,UART_TX_PROCH        ; switch to process using UART TX
    SCH_WAKEUP r16

    jmp sch_load

; uart_rx: Receive one byte, wait for it to arrive if necessary.
; If you are reading a stream of bytes, you may have to call uart_rx again
; within the time it takes to transfer one byte, to avoid a buffer
; overrun, as no software buffering is done.
; OUT   r16 = data byte
;       ZF = 1 on success, 0 on frame error/buffer overrun
uart_rx_wait:
    sts UART_RX_PROCL,yl        ; register our process as the one using
    sts UART_RX_PROCH,yh        ; the UART receiver
    cli
    sbi UCSRB,RXCIE             ; enable RX Complete interrupt
    call sch_wait               ; halt process until interrupt
uart_rx:
    sbis UCSRA,RXC              ; wait until the UART has new data for us
    rjmp uart_rx_wait
    in r16,UCSRA                ; r16 = status byte
    andi r16,0x18               ; check for frame error or buffer overrun
    in r16,UDR                  ; r16 = data byte (dummy value on error)
    sbi LED_PORT,0
    ret                         ; ZF will be 1 on success

; RX Complete interrupt handler.
; Switch to the process whole proc structure is pointed to by UART_RX_PROC
int_uart_rxc:
    cbi LED_PORT,0
    sbrc yh,7                   ; did we interrupt a process?
    rjmp int_uart_rxc_nosave    ; if yes, save its state

    call sch_saveregs
    in r0,SPL
    std y+0x06,r0
    in r0,SPH
    std y+0x07,r0

int_uart_rxc_nosave:
    cbi UCSRB,RXCIE             ; disable RX Complete interrupt

    lds yl,UART_RX_PROCL
    lds yh,UART_RX_PROCH        ; switch to process using UART RX
    SCH_WAKEUP r16

    jmp sch_load

